@@ -9,8 +9,9 @@ let () =
     (fun _ -> ())
     "test [options]"
 
+let () = if not (Sys.file_exists "out") then Sys.mkdir "out" 0o755
+
 let write fname s =
-  if not (Sys.file_exists "out") then Sys.mkdir "out" 0o755;
   let fname = "out/" ^ fname in
   let oc = open_out fname in
   output_string oc s;
@@ -41,6 +42,21 @@ let () =
   Printf.printf "- word size: %d\n%!" Sys.word_size;
   Printf.printf "\n%!"
 
+let () =
+  Printf.printf "## Testing basic functions\n\n%!";
+  assert (
+    try
+      ignore (Image.RGB8.Color.of_int 0xff00ff);
+      true
+    with _ -> false);
+  assert (
+    try
+      ignore
+        (Image.RGB8.Color.of_int
+           (Option.get (Int32.unsigned_to_int 0xaaff00ffl)));
+      false
+    with _ -> true)
+
 module A = Audio
 
 let () =
@@ -68,8 +84,7 @@ let () =
         A.U8.of_audio src 0 buf 0 len;
         A.U8.to_audio (Bytes.unsafe_to_string buf) 0 dst 0 len
       done;
-      assert (dst.(1).(len - 1) = 1.)
-    );
+      assert (dst.(1).(len - 1) = 1.));
   let src = A.make 2 len 1. in
   let buf = Bytes.create (A.S16LE.size 2 len) in
   let dst = A.create 2 len in
@@ -79,8 +94,7 @@ let () =
         A.S16LE.to_audio (Bytes.unsafe_to_string buf) 0 dst 0 len
       done;
       assert (dst.(1).(len - 1) = 1.);
-      assert (dst.(1).(len - 1) = 1.)
-    );
+      assert (dst.(1).(len - 1) = 1.));
   let src = A.make 2 len 1. in
   let buf = Bytes.create (A.S16BE.size 2 len) in
   let dst = A.create 2 len in
@@ -90,8 +104,7 @@ let () =
         A.S16BE.to_audio (Bytes.unsafe_to_string buf) 0 dst 0 len
       done;
       assert (dst.(1).(len - 1) = 1.);
-      assert (dst.(1).(len - 1) = 1.)
-    );
+      assert (dst.(1).(len - 1) = 1.));
   let src = A.make 2 len 1. in
   let buf = Bytes.create (A.S24LE.size 2 len) in
   let dst = A.create 2 len in
@@ -101,8 +114,7 @@ let () =
         A.S24LE.to_audio (Bytes.unsafe_to_string buf) 0 dst 0 len
       done;
       assert (dst.(1).(len - 1) = 1.);
-      assert (dst.(1).(len - 1) = 1.)
-    );
+      assert (dst.(1).(len - 1) = 1.));
   let src = A.make 2 len 1. in
   let buf = Bytes.create (A.S32LE.size 2 len) in
   let dst = A.create 2 len in
@@ -111,17 +123,47 @@ let () =
         A.S32LE.of_audio src 0 buf 0 len;
         A.S32LE.to_audio (Bytes.unsafe_to_string buf) 0 dst 0 len
       done;
-      assert (dst.(1).(len - 1) = 1.)
-    );
+      assert (dst.(1).(len - 1) = 1.));
   test "s16le with offset" (fun () ->
       let chans = 2 in
       let src = A.create chans len in
       let off = 21 in
       let buf = Bytes.create (off + A.S16LE.size chans len) in
-      A.S16LE.of_audio src 5 buf off (len-5);
-      A.S16LE.to_audio (Bytes.unsafe_to_string buf) off src 0 len
-    );
-  Printf.printf "\n"
+      A.S16LE.of_audio src 5 buf off (len - 5);
+      A.S16LE.to_audio (Bytes.unsafe_to_string buf) off src 0 len);
+  test "generate sine" (fun () ->
+      let chans = 2 in
+      let samplerate = 44100 in
+      let sl = new A.Mono.Generator.sine samplerate 440. in
+      let sr = new A.Mono.Generator.sine samplerate 880. in
+      let len = samplerate in
+      let buf = A.create chans len in
+      let f = new A.IO.Writer.to_wav_file chans samplerate "out/sine.wav" in
+      for _ = 0 to 9 do
+        sl#fill buf.(0) 0 len;
+        sr#fill buf.(1) 0 len;
+        f#write buf 0 len
+      done;
+      f#close);
+  time "compute replaygain" (fun () ->
+      let f = new A.IO.Reader.of_wav_file "out/sine.wav" in
+      let channels = f#channels in
+      let rg =
+        A.Analyze.ReplayGain.create ~channels ~samplerate:f#sample_rate
+      in
+      let loop = ref true in
+      let len = 1024 in
+      let buf = A.create channels len in
+      while !loop do
+        let n = f#read buf 0 len in
+        if n = 0 then loop := false
+        else A.Analyze.ReplayGain.process rg buf 0 len
+      done;
+      let gain = A.Analyze.ReplayGain.gain rg in
+      (* Printf.printf "[%.02f dB] %!" gain; *)
+      assert (abs_float (gain -. -14.56) < 0.05))
+
+let () = Printf.printf "\n"
 
 module I = Image
 
@@ -194,6 +236,25 @@ let () =
       let img = I.CanvasYUV420.make ~x:150 ~y:200 ~width:600 ~height:600 r in
       let img = I.CanvasYUV420.render img in
       write "canvas.bmp" (I.YUV420.to_BMP img));
+  test "scale grid proportionally" (fun () ->
+      let d = 100 in
+      let s = 10 in
+      let img = I.YUV420.create d d in
+      for j = 0 to d - 1 do
+        for i = 0 to d - 1 do
+          let c =
+            if ((i / s) + (j / s)) mod 2 = 0 then (0xff, 0xff, 0xff, 0xff)
+            else (0x00, 0x00, 0x00, 0xff)
+          in
+          I.YUV420.set_pixel_rgba img i j c
+        done
+      done;
+      let img =
+        I.CanvasYUV420.make img
+        |> I.CanvasYUV420.resize 640 362
+        |> I.CanvasYUV420.render
+      in
+      write "scale-grid.bmp" (I.YUV420.to_BMP img));
   test "scale canvas" (fun () ->
       let img = I.YUV420.create 1000 1000 in
       I.YUV420.gradient_uv img (0, 0) (0xff, 0) (0, 0xff);
@@ -235,6 +296,12 @@ let () =
       assert (not (I.YUV420.is_opaque img));
       I.YUV420.set_pixel_rgba img 10 10 (0, 0, 0, 0xff);
       assert (I.YUV420.is_opaque img));
+  test "add" (fun () ->
+      let a = I.YUV420.create 1280 480 in
+      let b = I.YUV420.create 640 480 in
+      let x = 1280 in
+      let y = 0 in
+      I.YUV420.add a ~x ~y b);
   time "many adds" (fun () ->
       let r = I.YUV420.create 500 500 in
       I.YUV420.fill r (I.Pixel.yuv_of_rgb (0xff, 0, 0));
@@ -281,8 +348,7 @@ let () =
       write "scale.bmp" (I.YUV420.to_BMP img2));
   test "font" (fun () ->
       let img = I.Bitmap.Font.render ~size:30 "Hello, world!\nHow are you?" in
-      write "hello-world.bmp" (I.YUV420.to_BMP (I.YUV420.of_bitmap img))
-    );
+      write "hello-world.bmp" (I.YUV420.to_BMP (I.YUV420.of_bitmap img)));
   time "sliding font" (fun () ->
       let width = 1280 in
       let height = 720 in
@@ -292,18 +358,16 @@ let () =
       let txt = I.YUV420.of_bitmap txt in
       let fname = "out/hello-world.avi" in
       let oc = open_out fname in
-      output_string oc (Video.AVI.Writer.header ~width ~height ~framerate:fps ());
+      output_string oc
+        (Video.AVI.Writer.header ~width ~height ~framerate:fps ());
       for i = 0 to duration * fps do
         let img = I.YUV420.create width height in
-        I.YUV420.fill img (I.Pixel.yuv_of_rgb (2*i,0,0xff));
-        I.YUV420.add txt ~x:(3*i) ~y:(2*i) img;
+        I.YUV420.fill img (I.Pixel.yuv_of_rgb (2 * i, 0, 0xff));
+        I.YUV420.add txt ~x:(3 * i) ~y:(2 * i) img;
         output_string oc (Video.AVI.Writer.Chunk.video_yuv420 img)
       done;
-      close_out oc
-    );
-  test "empty text" (fun () ->
-      ignore (I.Bitmap.Font.render ~size:20 "")
-    );
+      close_out oc);
+  test "empty text" (fun () -> ignore (I.Bitmap.Font.render ~size:20 ""));
   time "increasing saw" (fun () ->
       let width = 640 in
       let height = 360 in
@@ -313,7 +377,9 @@ let () =
       let channels = 2 in
       let samplerate = 44100 in
       let oc = open_out fname in
-      output_string oc (Video.AVI.Writer.header ~width ~height ~framerate:fps ~channels ~samplerate ());
+      output_string oc
+        (Video.AVI.Writer.header ~width ~height ~framerate:fps ~channels
+           ~samplerate ());
       let fmin = 20. in
       let fmax = 20000. in
       let duration = 20. in
@@ -324,24 +390,27 @@ let () =
       let buf = Audio.create channels (samplerate / fps) in
       let osc = ref (-1.) in
       while !t <= duration do
-        let f = fmin +. 2. ** (!t /. a) in
-        let txt = I.Bitmap.Font.render ~size:fontsize (Printf.sprintf "%.2f Hz" f) in
+        let f = fmin +. (2. ** (!t /. a)) in
+        let txt =
+          I.Bitmap.Font.render ~size:fontsize (Printf.sprintf "%.2f Hz" f)
+        in
         let txt = I.YUV420.of_bitmap txt in
         let img = I.YUV420.create width height in
         I.YUV420.blank img;
-        I.YUV420.add txt ~y:((height-fontsize)/2) img;
+        I.YUV420.add txt ~y:((height - fontsize) / 2) img;
         output_string oc (Video.AVI.Writer.Chunk.video_yuv420 img);
         for i = 0 to Audio.length buf - 1 do
           for c = 0 to Audio.channels buf - 1 do
             buf.(c).(i) <- !osc
           done;
-          osc := !osc +. 2. *. f /. float samplerate;
-          while !osc > 1. do osc := !osc -. 1.; done;
+          osc := !osc +. (2. *. f /. float samplerate);
+          while !osc > 1. do
+            osc := !osc -. 1.
+          done;
           t := !t +. dt
         done;
         output_string oc (Video.AVI.Writer.Chunk.audio_s16le buf)
       done;
-      close_out oc
-    )
+      close_out oc)
 
 let () = Gc.full_major ()
