@@ -117,6 +117,34 @@ module Sample = struct
     let x = max (-1.) x in
     let x = min 1. x in
     x
+
+  let iir a b =
+    let na = Array.length a in
+    let nb = Array.length b in
+    assert (a.(0) = 1.);
+    let x = Array.make nb 0. in
+    let y = Array.make na 0. in
+    let ka = ref 0 in
+    let kb = ref 0 in
+    fun x0 ->
+      let y0 = ref 0. in
+      x.(!kb) <- x0;
+      for i = 0 to nb - 1 do
+        y0 := !y0 +. (b.(i) *. x.((!kb + i) mod nb))
+      done;
+      for i = 1 to na - 1 do
+        y0 := !y0 -. (a.(i) *. y.((!ka + i) mod na))
+      done;
+      if na > 0 then y.(!ka) <- !y0;
+      let decr n k =
+        decr k;
+        if !k < 0 then k := !k + n
+      in
+      decr na ka;
+      decr nb kb;
+      !y0
+
+  let fir b = iir [||] b
 end
 
 module Mono = struct
@@ -161,6 +189,33 @@ module Mono = struct
     copy_to_ba buf ofs len ba;
     ba
 
+  external copy_from_int16_ba :
+    (int, Bigarray.int16_signed_elt, Bigarray.c_layout) Bigarray.Array1.t ->
+    float array ->
+    int ->
+    int ->
+    unit = "caml_mm_audio_copy_from_int16_ba"
+
+  external copy_to_int16_ba :
+    float array ->
+    int ->
+    int ->
+    (int, Bigarray.int16_signed_elt, Bigarray.c_layout) Bigarray.Array1.t ->
+    unit = "caml_mm_audio_copy_to_int16_ba"
+
+  let of_int16_ba buf =
+    let len = Bigarray.Array1.dim buf in
+    let dst = Array.create_float len in
+    copy_from_int16_ba buf dst 0 len;
+    dst
+
+  let to_int16_ba buf ofs len =
+    let ba =
+      Bigarray.Array1.create Bigarray.int16_signed Bigarray.c_layout len
+    in
+    copy_to_int16_ba buf ofs len ba;
+    ba
+
   let append b1 ofs1 len1 b2 ofs2 len2 =
     assert (length b1 - ofs1 >= len1);
     assert (length b2 - ofs2 >= len2);
@@ -210,9 +265,9 @@ module Mono = struct
       let s = Array.unsafe_get b (ofs + i) in
       Array.unsafe_set b (ofs + i)
         (if Float.is_nan s then 0.
-        else if s < -1. then -1.
-        else if 1. < s then 1.
-        else s)
+         else if s < -1. then -1.
+         else if 1. < s then 1.
+         else s)
     done
 
   let squares b ofs len =
@@ -957,6 +1012,17 @@ let copy_to_ba buf ofs len ba =
 let of_ba = Array.map Mono.of_ba
 let to_ba buf ofs len = Array.map (fun b -> Mono.to_ba b ofs len) buf
 
+let copy_from_int16_ba ba buf ofs len =
+  Array.iteri (fun i b -> Mono.copy_from_int16_ba ba.(i) b ofs len) buf
+
+let copy_to_int16_ba buf ofs len ba =
+  Array.iteri (fun i b -> Mono.copy_to_int16_ba buf.(i) ofs len b) ba
+
+let of_int16_ba = Array.map Mono.of_int16_ba
+
+let to_int16_ba buf ofs len =
+  Array.map (fun b -> Mono.to_int16_ba b ofs len) buf
+
 module U8 = struct
   let size channels samples = channels * samples
 
@@ -1178,6 +1244,195 @@ end
 module Analyze = struct
   let rms buf ofs len =
     Array.init (channels buf) (fun i -> Mono.Analyze.rms buf.(i) ofs len)
+
+  (* See https://github.com/FFmpeg/FFmpeg/blob/master/libavfilter/af_replaygain.c *)
+  (* See https://wiki.hydrogenaud.io/index.php?title=ReplayGain_specification *)
+
+  (** Replaygain computations. *)
+  module ReplayGain = struct
+    type t = {
+      channels : int;
+      mutable frame_pos : int;
+      frame_length : int;
+      prefilter : float array -> float array;
+      mutable peak : float;
+      mutable rms : float;
+      histogram : int array;
+    }
+
+    exception Not_supported
+
+    let histogram_slots = 12000
+
+    (** Create internal state. *)
+    let create =
+      let coeffs =
+        [
+          ( 48000,
+            ( [|
+                1.00000000000000;
+                -3.84664617118067;
+                7.81501653005538;
+                -11.34170355132042;
+                13.05504219327545;
+                -12.28759895145294;
+                9.48293806319790;
+                -5.87257861775999;
+                2.75465861874613;
+                -0.86984376593551;
+                0.13919314567432;
+              |],
+              [|
+                0.03857599435200;
+                -0.02160367184185;
+                -0.00123395316851;
+                -0.00009291677959;
+                -0.01655260341619;
+                0.02161526843274;
+                -0.02074045215285;
+                0.00594298065125;
+                0.00306428023191;
+                0.00012025322027;
+                0.00288463683916;
+              |],
+              [| 1.00000000000000; -1.97223372919527; 0.97261396931306 |],
+              [| 0.98621192462708; -1.97242384925416; 0.98621192462708 |] ) );
+          ( 44100,
+            ( [|
+                1.00000000000000;
+                -3.47845948550071;
+                6.36317777566148;
+                -8.54751527471874;
+                9.47693607801280;
+                -8.81498681370155;
+                6.85401540936998;
+                -4.39470996079559;
+                2.19611684890774;
+                -0.75104302451432;
+                0.13149317958808;
+              |],
+              [|
+                0.05418656406430;
+                -0.02911007808948;
+                -0.00848709379851;
+                -0.00851165645469;
+                -0.00834990904936;
+                0.02245293253339;
+                -0.02596338512915;
+                0.01624864962975;
+                -0.00240879051584;
+                0.00674613682247;
+                -0.00187763777362;
+              |],
+              [| 1.00000000000000; -1.96977855582618; 0.97022847566350 |],
+              [| 0.98500175787242; -1.97000351574484; 0.98500175787242 |] ) );
+          ( 22050,
+            ( [|
+                1.00000000000000;
+                -1.49858979367799;
+                0.87350271418188;
+                0.12205022308084;
+                -0.80774944671438;
+                0.47854794562326;
+                -0.12453458140019;
+                -0.04067510197014;
+                0.08333755284107;
+                -0.04237348025746;
+                0.02977207319925;
+              |],
+              [|
+                0.33642304856132;
+                -0.25572241425570;
+                -0.11828570177555;
+                0.11921148675203;
+                -0.07834489609479;
+                -0.00469977914380;
+                -0.00589500224440;
+                0.05724228140351;
+                0.00832043980773;
+                -0.01635381384540;
+                -0.01760176568150;
+              |],
+              [| 1.00000000000000; -1.94561023566527; 0.94705070426118 |],
+              [| 0.97316523498161; -1.94633046996323; 0.97316523498161 |] ) );
+        ]
+      in
+      fun ~channels ~samplerate ->
+        (* Frame length in samples (a frame is 50 ms). *)
+        let frame_length = samplerate * 50 / 1000 in
+        (* Coefficients of the Yulewalk and Butterworth filters. *)
+        let yule_a, yule_b, butter_a, butter_b =
+          match List.assoc_opt samplerate coeffs with
+            | Some c -> c
+            | None -> raise Not_supported
+        in
+        let yulewalk =
+          Array.init channels (fun _ -> Sample.iir yule_a yule_b)
+        in
+        let butterworth =
+          Array.init channels (fun _ -> Sample.iir butter_a butter_b)
+        in
+        let prefilter x =
+          Array.mapi (fun i x -> x |> yulewalk.(i) |> butterworth.(i)) x
+        in
+        {
+          channels;
+          frame_pos = 0;
+          frame_length;
+          prefilter;
+          peak = 0.;
+          rms = 0.;
+          histogram = Array.make histogram_slots 0;
+        }
+
+    (** Process a sample. *)
+    let process_sample rg x =
+      Array.iter
+        (fun x ->
+          let x = abs_float x in
+          if x > rg.peak then rg.peak <- x)
+        x;
+      let x = rg.prefilter x in
+      Array.iter (fun x -> rg.rms <- rg.rms +. (x *. x)) x;
+      rg.frame_pos <- rg.frame_pos + 1;
+      if rg.frame_pos >= rg.frame_length then (
+        (* Minimum value is about -100 dB for digital silence. The 90 dB
+           offset is to compensate for the normalized float range and 3 dB is
+           for stereo samples. *)
+        let rms =
+          (10. *. log10 (rg.rms /. float (rg.frame_length * rg.channels)))
+          +. 90.
+        in
+        let level =
+          int_of_float (100. *. rms) |> max 0 |> min (histogram_slots - 1)
+        in
+        rg.histogram.(level) <- rg.histogram.(level) + 1;
+        rg.rms <- 0.;
+        rg.frame_pos <- 0)
+
+    (** Process a buffer. *)
+    let process rg buf off len =
+      assert (channels buf = rg.channels);
+      for i = off to off + len - 1 do
+        let x = Array.init rg.channels (fun c -> buf.(c).(i)) in
+        process_sample rg x
+      done
+
+    (** Computed peak. *)
+    let peak rg = rg.peak
+
+    (** Compute gain. *)
+    let gain rg =
+      let windows = Array.fold_left ( + ) 0 rg.histogram in
+      let i = ref (histogram_slots - 1) in
+      let loud_count = ref 0 in
+      (* Find i below the top 5% *)
+      while !i > 0 && !loud_count * 20 < windows do
+        loud_count := !loud_count + rg.histogram.(!i);
+        decr i
+      done;
+      64.54 -. (float !i /. 100.) |> max (-24.) |> min 64.
+  end
 end
 
 module Effect = struct
@@ -1595,35 +1850,35 @@ module IO = struct
         method length = length
 
         initializer
-        if self#input 4 <> "RIFF" then
-          (* failwith "Bad header: \"RIFF\" not found"; *)
-          raise Invalid_file;
-        (* Ignore the file size *)
-        ignore (self#input 4);
-        if self#input 8 <> "WAVEfmt " then
-          (* failwith "Bad header: \"WAVEfmt \" not found"; *)
-          raise Invalid_file;
-        (* Now we always have the following uninteresting bytes:
-         * 0x10 0x00 0x00 0x00 0x01 0x00 *)
-        ignore (self#really_input 6);
-        channels <- self#input_short;
-        sample_rate <- self#input_int;
-        (* byt_per_sec *) ignore self#input_int;
-        (* byt_per_samp *) ignore self#input_short;
-        sample_size <- self#input_short;
-
-        let section = self#really_input 4 in
-        if section <> "data" then (
-          if section = "INFO" then
-            (* failwith "Valid wav file but unread"; *)
+          if self#input 4 <> "RIFF" then
+            (* failwith "Bad header: \"RIFF\" not found"; *)
             raise Invalid_file;
-          (* failwith "Bad header : string \"data\" not found" *)
-          raise Invalid_file);
+          (* Ignore the file size *)
+          ignore (self#input 4);
+          if self#input 8 <> "WAVEfmt " then
+            (* failwith "Bad header: \"WAVEfmt \" not found"; *)
+            raise Invalid_file;
+          (* Now we always have the following uninteresting bytes:
+           * 0x10 0x00 0x00 0x00 0x01 0x00 *)
+          ignore (self#really_input 6);
+          channels <- self#input_short;
+          sample_rate <- self#input_int;
+          (* byt_per_sec *) ignore self#input_int;
+          (* byt_per_samp *) ignore self#input_short;
+          sample_size <- self#input_short;
 
-        let len_dat = self#input_int in
-        data_offset <- self#stream_cur_pos;
-        bytes_per_sample <- sample_size / 8 * channels;
-        length <- len_dat / bytes_per_sample
+          let section = self#really_input 4 in
+          if section <> "data" then (
+            if section = "INFO" then
+              (* failwith "Valid wav file but unread"; *)
+              raise Invalid_file;
+            (* failwith "Bad header : string \"data\" not found" *)
+            raise Invalid_file);
+
+          let len_dat = self#input_int in
+          data_offset <- self#stream_cur_pos;
+          bytes_per_sample <- sample_size / 8 * channels;
+          length <- len_dat / bytes_per_sample
 
         method read (buf : buffer) ofs len =
           let sbuflen = len * channels * 2 in
@@ -1632,9 +1887,9 @@ module IO = struct
           let len = sbuflen / (channels * 2) in
           begin
             match sample_size with
-            | 16 -> S16LE.to_audio sbuf 0 buf ofs len
-            | 8 -> U8.to_audio sbuf 0 buf ofs len
-            | _ -> assert false
+              | 16 -> S16LE.to_audio sbuf 0 buf ofs len
+              | 8 -> U8.to_audio sbuf 0 buf ofs len
+              | _ -> assert false
           end;
           len
 
@@ -1676,27 +1931,28 @@ module IO = struct
         method virtual private sample_rate : int
 
         initializer
-        let bits_per_sample = 16 in
-        (* RIFF *)
-        self#output "RIFF";
-        self#output_int 0;
-        self#output "WAVE";
+          let bits_per_sample = 16 in
+          (* RIFF *)
+          self#output "RIFF";
+          self#output_int 0;
+          self#output "WAVE";
 
-        (* Format *)
-        self#output "fmt ";
-        self#output_int 16;
-        self#output_short 1;
-        self#output_short self#channels;
-        self#output_int self#sample_rate;
-        self#output_int (self#sample_rate * self#channels * bits_per_sample / 8);
-        self#output_short (self#channels * bits_per_sample / 8);
-        self#output_short bits_per_sample;
+          (* Format *)
+          self#output "fmt ";
+          self#output_int 16;
+          self#output_short 1;
+          self#output_short self#channels;
+          self#output_int self#sample_rate;
+          self#output_int
+            (self#sample_rate * self#channels * bits_per_sample / 8);
+          self#output_short (self#channels * bits_per_sample / 8);
+          self#output_short bits_per_sample;
 
-        (* Data *)
-        self#output "data";
-        (* size of the data, to be updated afterwards *)
-        self#output_short 0xffff;
-        self#output_short 0xffff
+          (* Data *)
+          self#output "data";
+          (* size of the data, to be updated afterwards *)
+          self#output_short 0xffff;
+          self#output_short 0xffff
 
         val mutable datalen = 0
 
@@ -1736,8 +1992,8 @@ module IO = struct
         method virtual io_write : buffer -> unit
 
         initializer
-        assert (fill_duration <= max_duration);
-        assert (drop_duration <= max_duration)
+          assert (fill_duration <= max_duration);
+          assert (drop_duration <= max_duration)
 
         val rb = Ringbuffer.create channels max_duration
 
