@@ -164,8 +164,8 @@ CAMLprim value caml_yuv420_scale(value _src, value _dst) {
     }
   for (j = 0; j < dst.height / 2; j++)
     for (i = 0; i < dst.width / 2; i++) {
-      is = i * src.width / dst.width;
-      js = j * src.height / dst.height;
+      is = i * (src.width / 2) / (dst.width / 2);
+      js = j * (src.height / 2) / (dst.height / 2);
       U2(dst, i, j) = U2(src, is, js);
       V2(dst, i, j) = V2(src, is, js);
     }
@@ -235,6 +235,9 @@ CAMLprim value caml_yuv420_add(value _src, value _x, value _y, value _dst) {
   int ja = max(y, 0);
   int jb = min(y + src.height, dst.height);
 
+  if (!(ia < ib))
+    CAMLreturn(Val_unit);
+
   caml_enter_blocking_section();
   if (src.alpha == NULL) {
     int il = ib - ia;
@@ -408,17 +411,25 @@ CAMLprim value caml_yuv_greyscale(value img) {
 }
 
 #define PIXEL_PRECISON 0x10000
-CAMLprim value caml_yuv_scale_alpha(value img, value _a) {
-  CAMLparam2(img, _a);
+CAMLprim value caml_yuv_scale_alpha(value img, value _c) {
+  CAMLparam2(img, _c);
   yuv420 yuv;
   yuv420_of_value(&yuv, img);
-  int a = Double_val(_a) * PIXEL_PRECISON;
+  int c = Double_val(_c) * PIXEL_PRECISON;
+  int amax = CLIP(0xff * c / PIXEL_PRECISON);
   int i, j;
 
   caml_enter_blocking_section();
   for (j = 0; j < yuv.height; j++)
-    for (i = 0; i < yuv.width; i++)
-      A(yuv, i, j) = CLIP(A(yuv, i, j) * a / PIXEL_PRECISON);
+    for (i = 0; i < yuv.width; i++) {
+      int a = A(yuv, i, j);
+      if (a != 0) {
+        if (a == 0xff)
+          A(yuv, i, j) = amax;
+        else
+          A(yuv, i, j) = CLIP(a * c / PIXEL_PRECISON);
+      }
+    }
   caml_leave_blocking_section();
 
   CAMLreturn(Val_unit);
